// Copyright (c) 2011 Rane Corporation. All Rights Reserved.
// $Id$
//
// Performs a variety of non-streaming tests on the audio device.

#include <stdio.h>

#ifdef WIN32
#include "windows.h"
#else
#include <unistd.h>
#include <stdlib.h>
#include <IOKit/IOKitLib.h>

#endif

#include "SixtyTwoApi.h"
#include "SixtyTwoApiTest.h"

/// Quick and dirty testing macro expanded inline; verifies that the API function
/// returned zero for success, printing a message on error.
/// Increments local vars for failure and success counts for summary display.
#define mTest(description, resultExpr)              \
{                                                   \
int resultIntPrivate = resultExpr;                  \
if (resultIntPrivate != RH62_RESULT_SUCCESS)        \
{                                                   \
printf("****** FAILED: %s (%s) *******\n",          \
description,                                        \
::GetErrorCodeDesc(resultIntPrivate));              \
++failureCount;                                     \
}                                                   \
else                                                \
{                                                   \
++successCount;                                     \
}                                                   \
}

/// Same as mTest, but takes a boolean "success" indicator instead of an MP
/// error code constant.
#define mTestBool(successBool, description)         \
mTest(description, (successBool ?                   \
RH62_RESULT_SUCCESS :                               \
RH62_RESULT_UNSPECIFIED_ERROR))

/// Retrieve the string name of a given RH62_RESULT_xyz API call result code.
const char* GetErrorCodeDesc(int errorCode)
{
   return
   errorCode == RH62_RESULT_SUCCESS ? "RH62_RESULT_SUCCESS" :
   errorCode == RH62_RESULT_UNSPECIFIED_ERROR ? "RH62_RESULT_UNSPECIFIED_ERROR" :
   errorCode == RH62_RESULT_UNIMPLEMENTED ? "RH62_RESULT_UNIMPLEMENTED" :
   errorCode == RH62_RESULT_INVALID_HANDLE ? "RH62_RESULT_INVALID_HANDLE" :
   errorCode == RH62_RESULT_COMMUNICATION_ERROR ? "RH62_RESULT_COMMUNICATION_ERROR" :
   errorCode == RH62_RESULT_INVALID_DRIVER_STATE ? "RH62_RESULT_INVALID_DRIVER_STATE" :
   errorCode == RH62_RESULT_DRIVER_ERROR ? "RH62_RESULT_DRIVER_ERROR" :
   errorCode == RH62_RESULT_ASYNC_CALL ? "RH62_RESULT_ASYNC_CALL" :
   errorCode == RH62_RESULT_INVALID_PARAMETER ? "RH62_RESULT_INVALID_PARAMETER" :
   errorCode == RH62_RESULT_INSUFFICIENT_SPACE ? "RH62_RESULT_INSUFFICIENT_SPACE" :
   errorCode == RH62_RESULT_EMPTY_PACKET ? "RH62_RESULT_EMPTY_PACKET" :
   errorCode == RH62_RESULT_TYPE_MISMATCH ? "RH62_RESULT_TYPE_MISMATCH" :
   "(unknown error code)";
}

/// Delay execution for specified duration.
void SleepForMillisecs(int millisecs)
{
#ifdef WIN32
   ::Sleep(millisecs);
#else
   usleep(millisecs * 1000);
#endif
}

/// App-defined callback to report firmware update progress
void RH62_API_CALLBACK_CONV ReportFirmwareProgress(RH62_firmware_status_data_t* firmwareStatus)
{
   printf("...%d", firmwareStatus->progress);
}

/// App-defined callback to report HIDI messages
void RH62_API_CALLBACK_CONV HandleHidiMessage(RH62_hidimsg_data_t* data)
{
   if (data->length > RH62_MAX_HIDI_MESSAGE_LENGTH)
   {
      printf("ERROR -- HIDI message length greater than max allowed: %d > %d.",
             data->length, RH62_MAX_HIDI_MESSAGE_LENGTH);
      getchar();
      exit(1);

   }
   bool isAllF8s = true;
   for (int i = 0; i < data->length; ++i)
   {
      if (data->msg[i] != 0xf8)
      {
         isAllF8s = false;
         break;
      }
   }
   if (!isAllF8s)
   {
      printf("Got HIDI msg: ");
      for (int i = 0; i < data->length; ++i)
      {
         printf("%s%02x", (i > 0 ? "," : ""), data->msg[i]);
      }
      printf("\n");
   }
}

/// App-defined callback to report port connect messages
void RH62_API_CALLBACK_CONV HandleUsbPortStatusMessage(RH62_usb_port_status_t* data)
{
   printf("Got usb port status message showing \"this\" port: %s; \"other\" port: %s\n",
          (RH62_USB_PORT_A == data->thisPort ? "A" :
           RH62_USB_PORT_B == data->thisPort ? "B" :
           "(disconnected)"),
          (RH62_USB_PORT_A == data->otherPort ? "A" :
           RH62_USB_PORT_B == data->otherPort ? "B" :
           "(disconnected)")
          );
}

/// This should be invoked from the entry point of the app; it runs through
/// multiple tests of the Rane audio control API.
/// To support future flexibility, argc and argv should be simply passed through.
int RunAllSixtyTwoTests(int argc, _TCHAR* const argv[])
{
   int successCount = 0;
   int failureCount = 0;

   int resultInt = 0;
   void* driverHandle;

   printf("------------------------------------------\n");
   printf("Starting MP audio control API test...\n\n");

   while (true)
   {
      // open a handle to the driver
      driverHandle = ::RH62_open();
      int openResult = driverHandle ? RH62_RESULT_SUCCESS : RH62_RESULT_INVALID_HANDLE;
      mTest("RH62_open", openResult);
      if (driverHandle)
      {
         break; // success; no need to retry
      }
      printf("ERROR -- Failed to find a device to connect to.\nPress \"r\" and "
             "ENTER to Retry opening the device, or just ENTER to skip.\n");
      if ('r' == ::getchar())
      {
         ::getchar(); // consume the CR as well
         continue;
      }
   }

   // Retrieves the operating firmware version information.
   RH62_version_t firmwareVersion = {0};
   mTest("RH62_get_firmware_version",
         ::RH62_get_firmware_version(driverHandle, &firmwareVersion));
   printf("Firmware operating version: %d.%d.%d.%d\n",
          firmwareVersion.major,
          firmwareVersion.minor,
          firmwareVersion.sub,
          firmwareVersion.build);

   /*
   // Enable this temporarily to test flush HIDI queue function.
   
   ::RH62_set_hidimsg_callback(driverHandle, HandleHidiMessage);
   RH62_hidimsg_data_t hididata;
   hididata.length = 3;
   hididata.msg[0] = 0xB0;
   hididata.msg[1] = 22;
   hididata.msg[2] = 0;
   RH62_send_hidimsg(driverHandle, &hididata);
   RH62_flush_hidi_queue(driverHandle);
   getchar();
   return 0;
   */

   // Retrieves the embedded firmware update version information.
   mTest("RH62_get_firmware_embedded_version",
         ::RH62_get_firmware_embedded_version(
                                            driverHandle,
                                            &firmwareVersion));
   printf("Firmware update embedded version: %d.%d.%d.%d\n",
          firmwareVersion.major,
          firmwareVersion.minor,
          firmwareVersion.sub,
          firmwareVersion.build);

   // test boot device status
   int isBootDeviceInt = 0;
   mTest("RH62_is_boot_device", ::RH62_is_boot_device(driverHandle, &isBootDeviceInt));
   printf("Is boot device: %d\n", isBootDeviceInt);

   if (isBootDeviceInt)
   {
      printf("WARNING -- since boot device detected, firmware update is strongly recommended.\n");
   }

   printf("------------------------------------------------------------------------------\n");
   printf("Pausing for optional firmware update testing (from embedded image).\n"
          "Press \"u\" and ENTER to start firmware update, or just ENTER to skip.\n");
   if ('u' == ::getchar())
   {
      ::getchar(); // eat the CR
      mTest("RH62_set_update_firmware_from_embedded",
            ::RH62_set_update_firmware_from_embedded(driverHandle,
                                                   ReportFirmwareProgress));
      printf("\nDone testing firmware update.\n");

      mTest("RH62_is_boot_device", ::RH62_is_boot_device(driverHandle, &isBootDeviceInt));
      printf("Is boot device: %d\n", isBootDeviceInt);

      mTestBool(!isBootDeviceInt, "still boot device");
      if (isBootDeviceInt)
      {
         printf("WARNING -- still boot device; firmware update must have failed.\n");
      }
   }
   else
   {
      printf("Skipping firmware update.\n");
   }
   printf("Continuing tests.\n\n");

   // Retrieves the Vendor ID of the device.
   mTest("RH62_get_vendor_id", ::RH62_get_vendor_id(driverHandle, &resultInt));
   printf("USB VID: 0x%x\n", resultInt);

   // Retrieves the Product ID of the device.
   mTest("RH62_get_product_id", ::RH62_get_product_id(driverHandle, &resultInt));
   printf("USB PID: 0x%x\n", resultInt);

   // Retrieves API version information.
   RH62_version_t apiVersion = {0};
   mTest("RH62_get_api_version", ::RH62_get_api_version(&apiVersion));
   printf("API version: %d.%d.%d.%d\n",
          apiVersion.major,
          apiVersion.minor,
          apiVersion.sub,
          apiVersion.build);

   // Retrieves the usb port status of the device.
   RH62_usb_port_status_t usbPortStatus;
   mTest("RH62_get_usb_port_status", ::RH62_get_usb_port_status(driverHandle, &usbPortStatus));
   printf("Device usb port status: this=%d, other=%d\n",
          usbPortStatus.thisPort,
          usbPortStatus.otherPort);

   // setup connect/disconnect notification callbacks
   mTest("RH62_set_usb_port_status_callback",
         RH62_set_usb_port_status_callback(driverHandle, HandleUsbPortStatusMessage));

   // Retrieves the driver version string.
   RH62_version_t driverVersion = {0};
   mTest("RH62_get_driver_version",
         ::RH62_get_driver_version(driverHandle, &driverVersion));
   printf("Driver version: %d.%d.%d.%d\n",
          driverVersion.major,
          driverVersion.minor,
          driverVersion.sub,
          driverVersion.build);

   // Requests a driver timer challenge/response cycle.
   // Below test data per Roy Gill, 10/12/2009
   unsigned char kSendChallenge[RH62_DRIVER_TIMER_LENGTH] =
   {0xba, 0x1e, 0x6f, 0x78, 0xca, 0x2e, 0x65, 0xb9};
   unsigned char kExpectedResponse[RH62_DRIVER_TIMER_LENGTH] =
   {0x5e, 0xcb, 0x0d, 0x63, 0xd0, 0xa7, 0xf9, 0x08};
   unsigned char actualResponse[RH62_DRIVER_TIMER_LENGTH] = {0};
   mTest("RH62_driver_timer",
         ::RH62_driver_timer(driverHandle,
                           kSendChallenge,
                           RH62_DRIVER_TIMER_LENGTH,
                           actualResponse,
                           RH62_DRIVER_TIMER_LENGTH));
   printf("Timer challenge: 0x%llx; response: 0x%llx\n"
          "(Expected response: 0x%llx)\n",
          *((long long*)kSendChallenge),
          *((long long*)actualResponse),
          *((long long*)kExpectedResponse));
   for (int i = 0; i < RH62_DRIVER_TIMER_LENGTH; ++i)
   {
      if (actualResponse[i] != kExpectedResponse[i])
      {
         mTest("Compare driver timer challenge/response round trip result",
               RH62_RESULT_UNSPECIFIED_ERROR);
         break;
      }
   }


   printf("------------------------------------------------------------------------------\n");
   printf("Pausing for optional disconnect testing; feel free to disconnect and reconnect\n"
          "the USB cable, wait several seconds, then press ENTER to continue.\n"
          "If the rest of the test session fails, then reconnect is likely broken.\n");
   getchar();
   printf("Resuming.\n\n");

   // test control value handling
   mTest("RH62_set_hidimsg_callback",
         ::RH62_set_hidimsg_callback(driverHandle, HandleHidiMessage));
   for (int i = 0; i < 1; ++i)
   {
      printf("Requesting all control values to be sent to callback; should see\n");
      printf("activity printed onscreen afterwards...\n");
      mTest("RH62_get_all_controls",
            ::RH62_get_all_controls(driverHandle));
      printf("Waiting for callbacks...  ");
      ::SleepForMillisecs(5000);
      printf("...done waiting.\n");
   }

   mTest("RH62_send_control_change Input Select 1 to 0x03 (line)",
         ::RH62_send_control_change(driverHandle, 0x70, 0x03));
   ::SleepForMillisecs(500);

   // test setting control values
   RH62_hidimsg_data_t setControlsPacket = {0};
   setControlsPacket.length = 3;
   setControlsPacket.msg[0] = RH62_MIDI_STATUS_CONTROL_CHANGE;
   setControlsPacket.msg[1] = RH62_PARAM_A1_PHONO_SENSITIVITY;
   setControlsPacket.msg[2] = 0x05;
   mTest("RH62_send_hidimsg",
         ::RH62_send_hidimsg(
                           driverHandle,
                           &setControlsPacket));

   mTest("RH62_send_control_change",
         ::RH62_send_control_change(
                                  driverHandle,
                                  RH62_PARAM_A1_PHONO_SENSITIVITY,
                                  5));

   printf("Sending system exclusive command to set control point.\n");
   const int kSysExCommandSize = 2;
   unsigned char kSysExCommand[kSysExCommandSize] = {
      RH62_SYSEX_COMMAND_GET_ALL,
   usbPortStatus.thisPort};
   mTest("RH62_send_sys_ex",
         ::RH62_send_sys_ex(
                          driverHandle,
                          kSysExCommand,
                          kSysExCommandSize));
   ::SleepForMillisecs(1000); // give time to report HIDI notification

   // Retrieves the old buffer size for restoration later
   int oldBufferMillisecs = 0;
   mTest("RH62_get_buffer_millisecs",
         ::RH62_get_buffer_millisecs(
                                   driverHandle,
                                   &oldBufferMillisecs));
   printf("Read current buffer millisecs: %d\n", oldBufferMillisecs);

   // Sets the current buffer size
   int testBufferMillisecs = ((9 == oldBufferMillisecs) ? 10 : 9);
   mTest("RH62_set_buffer_millisecs",
         ::RH62_set_buffer_millisecs(
                                   driverHandle,
                                   testBufferMillisecs));
   printf("Set buffer millisecs to: %d\n", testBufferMillisecs);

   // Retrieves the current buffer millisecs
   mTest("RH62_get_buffer_millisecs",
         ::RH62_get_buffer_millisecs(
                                   driverHandle,
                                   &resultInt));
   printf("Read new buffer millisecs: %d\n", resultInt);

   mTestBool("Buffer millisecs round trip check", resultInt == testBufferMillisecs);

   // restore old buffer millisecs
   printf("Restoring old buffer millisecs\n");
   mTest("RH62_get_buffer_millisecs",
         ::RH62_set_buffer_millisecs(
                                   driverHandle,
                                   oldBufferMillisecs));

   RH62_send_control_change(
    driverHandle,
    RH62_PARAM_ANLG_INSERT_SENSITIVITY,
    0x00);
   printf("Just now set analog sensitivity to +4dBu... press Enter");
   getchar();
   RH62_send_control_change(
    driverHandle,
    RH62_PARAM_ANLG_INSERT_SENSITIVITY,
    0x22);
   printf("Just now set analog sensitivity to +4dBu... press Enter");
   getchar();
   RH62_send_control_change(
    driverHandle,
    RH62_PARAM_ANLG_INSERT_SENSITIVITY,
    0x3f);
   printf("Just now set analog sensitivity to +4dBu... press Enter");
   getchar();

   RH62_send_control_change(
    driverHandle,
    RH62_PARAM_ANLG_INSERT_SENSITIVITY,
    0x40);
   printf("Just now set analog sensitivity to -104dBV... press Enter");
   getchar();
   RH62_send_control_change(
    driverHandle,
    RH62_PARAM_ANLG_INSERT_SENSITIVITY,
    0x66);
   printf("Just now set analog sensitivity to -104dBV... press Enter");
   getchar();
   RH62_send_control_change(
    driverHandle,
    RH62_PARAM_ANLG_INSERT_SENSITIVITY,
    0x7f);
   printf("Just now set analog sensitivity to -104dBV... press Enter");
   getchar();

   printf("\n------------------------------------------------------------------------------\n");
   printf("Pausing for optional alternate product ID test.\n"
          "Press \"a\" and ENTER to switch to alternate product ID, or just ENTER to skip.\n");

   if ('a' == ::getchar())
   {
      ::getchar(); // eat the CR
      mTest("RH62_set_alt_product_id",
            ::RH62_set_alt_product_id(driverHandle));
      printf("Done testing alternate product ID.  NOTE: Device must be rebooted before additional testing\n\n");
   }
   else
   {
      printf("Skipping alternate product id.\n");
   }
   printf("Continuing tests.\n\n");


   // Close device handle.
   printf("Closing driver handle.\n");
   ::RH62_close(driverHandle);

   // Report results to user, and tell user we're waiting for ENTER to continue
   bool success = 0 == failureCount;
   const char* resultDesc = success ? "Success" : "FAILED";
   printf("\n\nResults:\n%s - %d of %d operations succeeded.\n"
          "Press ENTER to continue.\n",
          resultDesc, // "success"/ "failed"
          successCount,
          successCount + failureCount);
   getchar();

   return failureCount;
}

